// -*- C++ -*-
/*!
 * @file  pa10vel.cpp * @brief velocity controlled pa10 arm * $Date$ 
 *
 * $Id$ 
 */
#include "pa10vel.h"

// Module specification
// <rtc-template block="module_spec">
static const char* pa10vel_spec[] =
  {
    "implementation_id", "pa10vel",
    "type_name",         "pa10vel",
    "description",       "velocity controlled pa10 arm",
    "version",           "1.0",
    "vendor",            "Takashi Suehiro, AIST",
    "category",          "vel_7dof",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    ""
  };
// </rtc-template>

pa10vel::pa10vel(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_velIn("vel", m_vel),
    m_thOut("th", m_th)

    // </rtc-template>
{
  m_velcb= new VelCB(this);
  m_velIn.setOnWrite(m_velcb);
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("vel", m_velIn);

  // Set OutPort buffer
  registerOutPort("th", m_thOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

}

pa10vel::~pa10vel()
{
}


RTC::ReturnCode_t pa10vel::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable

  // </rtc-template>
  int error;
  int i;

  m_th.data.length(7);
  for(i=0;i<7;i++) {
    ref_vel[i]=0.0;
  }

  // alloc
  pa10    =  pa10_info_alloc();
  if(pa10 == NULL) {
    printf("pa10 alloc error \n");
    return RTC::RTC_ERROR;
  }
  pa10_st =  pa10_state_alloc();
  if(pa10_st == NULL) {
    printf("pa10_st alloc error \n");
    return RTC::RTC_ERROR;
  }

  // set initial parameter
  pa10->Id        = 0;            // ID
  pa10->Type      = pa10_info::PA_10A_ARM;   // PA_10A_ARM
  pa10->TipLength = 0.100;        // [m]

  // PA-10 initialized
  error = pa10_open(pa10);
  if(error != 0){
    printf("PA10 initialized error %d \n", error);
    return RTC::RTC_ERROR;
  }

  return RTC::RTC_OK;
}


RTC::ReturnCode_t pa10vel::onFinalize()
{
  int error;

  // arcnet board close
  error = pa10_close(pa10);
  if(error != 0){
    printf("PA10 close error %d \n", error);
    return RTC::RTC_ERROR;
  }

  // memory free
  pa10_info_free(pa10);
  pa10_state_free(pa10_st);
  printf("program finish \n");
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t pa10vel::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t pa10vel::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t pa10vel::onActivated(RTC::UniqueId ec_id)
{
  int error;
  int i;
  timeval time_val;

  // anyway clear alarm
  error = pa10_alarm_clear(pa10);
  if(error != 0){
    printf("PA10 alarm clear error %d \n", error);
    return RTC::RTC_ERROR;
  }

  //  tim.new_time(n_tm.sec, n_tm.nsec);
  gettimeofday(&time_val, NULL);
  in_tm.sec = time_val.tv_sec;
  in_tm.nsec = time_val.tv_usec * 1000;

  for(i=0;i<7;i++) {
    ref_vel[i]=0.0;
  }

  // start, only one time
  error = pa10_start(pa10);
  if(error != 0){
    printf("PA10 start error : %d \n", error);
    return RTC::RTC_ERROR;
  }
  // send servo data
  error = pa10_joint_velocity_control(pa10, ref_vel);
  if(error != 0){
    printf("joint velocity control error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  error = pa10_get_current_data(pa10, pa10_st);
  if(error != 0){
    printf("PA10 get_current error : %d \n", error);
    return RTC::RTC_ERROR;
  }
  for(i=0;i<7;i++) {
    m_th.data[i]=pa10_st->CurJoint[i];
  }
  m_thOut.write();
  //  usleep(1000);

  // send servo data AGAIN. this must be at the end of the function
  error = pa10_joint_velocity_control(pa10, ref_vel);
  if(error != 0){
    printf("joint velocity control error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  return RTC::RTC_OK;
}


RTC::ReturnCode_t pa10vel::onDeactivated(RTC::UniqueId ec_id)
{

  int error;
  int i;

  for(i=0;i<7;i++) {
    ref_vel[i]=0.0;
  }

  // send servo data
  error = pa10_joint_velocity_control(pa10, ref_vel);
  if(error != 0){
    printf("joint velocity control error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  // sending data stop
  error = pa10_stop(pa10);
  if(error != 0){
    printf("PA10 stop error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  return RTC::RTC_OK;
}

//-------------------------------------------------------------
// helper functions
//-------------------------------------------------------------
double calc_tm_dif(RTC::Time tm1, RTC::Time tm2) {
  return (double)(tm1.sec - tm2.sec)
    + (double)(1000000000 + tm1.nsec - tm2.nsec)/ 1000000000.0 - 1.0;
}
//
double calc_decay(double d_ftm, double l_ftm, double t_cnst) {
  if(d_ftm > l_ftm) {
    return exp((l_ftm - d_ftm)/ t_cnst);
  } else {
    return 1.0;
  }
}
//--------------------------------------------------------------
RTC::ReturnCode_t pa10vel::onExecute(RTC::UniqueId ec_id)
{
  int error;
  int i;

  RTC::Time n_tm;
  timeval time_val;
  double d_tm, d_fact;
  double vel[7];

  gettimeofday(&time_val, NULL);
  n_tm.sec = time_val.tv_sec;
  n_tm.nsec = time_val.tv_usec * 1000;

  d_tm = calc_tm_dif(n_tm, in_tm);  // delta time
  //  std::cout << d_tm;
  //  d_fact = calc_decay(d_tm, 0.1, 1.0);  // apply decay
  d_fact = calc_decay(d_tm, 0.1, 0.5);  // apply decay
  for(i=0;i<7;i++) {
    vel[i]=d_fact*ref_vel[i];
  }
  // send servo data
  error = pa10_joint_velocity_control(pa10, vel);
  if(error != 0){
    printf("joint velocity control error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  error = pa10_get_current_data(pa10, pa10_st);
  if(error != 0){
    printf("PA10 get_current error : %d \n", error);
    return RTC::RTC_ERROR;
  }
  for(i=0;i<7;i++) {
    m_th.data[i]=pa10_st->CurJoint[i];
  }
  m_thOut.write();
  //  usleep(10000);
  usleep(1000);
  return RTC::RTC_OK;
}


RTC::ReturnCode_t pa10vel::onAborting(RTC::UniqueId ec_id)
{
  int error;
  int i;

  for(i=0;i<7;i++) {
    ref_vel[i]=0.0;
  }

  // send servo data
  error = pa10_joint_velocity_control(pa10, ref_vel);
  if(error != 0){
    printf("joint velocity control error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  // sending data stop
  error = pa10_stop(pa10);
  if(error != 0){
    printf("PA10 stop error: %d\n", error);
    return RTC::RTC_ERROR;
  }
  return RTC::RTC_OK;
}

RTC::ReturnCode_t pa10vel::onError(RTC::UniqueId ec_id)
{
  //int error;
  //int i;

  //  for(i=0;i<7;i++) {
  //     ref_vel[i]=0.0;
  //}

  // send servo data
  //error = pa10_joint_velocity_control(pa10, ref_vel);
  //if(error != 0){
  //  printf("joint velocity control error: %d\n", error);
  //  return RTC::RTC_ERROR;
  //}
  // sending data stop
  //error = pa10_stop(pa10);
  //if(error != 0){
  //  printf("PA10 stop error: %d\n", error);
  //  return RTC::RTC_ERROR;
  //}
  usleep(10000);
  return RTC::RTC_OK;
}


RTC::ReturnCode_t pa10vel::onReset(RTC::UniqueId ec_id)
{
  int error;

  // memory free
  //  arcnet board close
    error = pa10_close(pa10);
    if(error != 0){
      printf("PA10 close error %d \n", error);
    }
  //  pa10_info_free(pa10);
  //  pa10_state_free(pa10_st);
  //  usleep(1000000);

  // alloc
  //  pa10    =  pa10_info_alloc();
  //  if(pa10 == NULL) {
  //    printf("pa10 alloc error \n");
  //    return RTC::RTC_ERROR;
  //  }
  //  pa10_st =  pa10_state_alloc();
  //  if(pa10_st == NULL) {
  //    printf("pa10_st alloc error \n");
  //    return RTC::RTC_ERROR;
  //  }

  // set initial parameter
  //  pa10->Id        = 0;            // ID
  //  pa10->Type      = pa10_info::PA_10A_ARM;   // PA_10A_ARM
  //  pa10->TipLength = 0.100;        // [m]

  //   PA-10 open
    error = pa10_open(pa10);
    if(error != 0){
      printf("PA10 open error %d \n", error);
      return RTC::RTC_ERROR;
    }
    usleep(500000);
  // alarm clear
  error = pa10_alarm_clear(pa10);
  if(error != 0){
    printf("PA10 alarm clear error %d \n", error);
    return RTC::RTC_ERROR;
  }
  usleep(500000);
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t pa10vel::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t pa10vel::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


extern "C"
{
 
  void pa10velInit(RTC::Manager* manager)
  {
    RTC::Properties profile(pa10vel_spec);
    manager->registerFactory(profile,
                             RTC::Create<pa10vel>,
                             RTC::Delete<pa10vel>);
  }
  
};



